---
hide-toc: true
---

# The Typon compiler

```{toctree} 
:hidden:
:caption: Getting started

gettingstarted/cli
```

```{toctree} 
:hidden:
:caption: Language

language/differences
language/typesystem
language/concurrency
```

```{toctree} 
:hidden:
:caption: Interoperability

interoperability/cpython
```

## What's this?

Typon is a work-in-progress **Python-to-C++ compiler**.

Untyped, idiomatic Python code can be compiled to C++, which can then be compiled to native code using a C++ compiler.

## Supported features

Typon's goal is to support as big a subset of Python as possible while still being able to process *idiomatic* Python
code. This means:

- **type inference and type checking**: a successful Typon build means no type errors in the C++ code
- **support for nested functions and closures**
- **automatic memory management**: objects are garbage-collected using reference counting
- **interoperability with native Python**: Typon can call Python functions and use Python objects (CPython will then be
  used, only for that use)
- **first-class types and functions**: Typon supports passing and storing types and functions as values

## What's new?

You may be thinking "why yet another Python compiler?". Typon was born out of the observation that none of the existing
Python compilers focused on concurrency and asynchronous programming.

## Comparison

Typon is not the first Python compiler, many other projects have similar goals. Here is a comparison with some of them:

|                               | Cython              | Nuitka                                       | Codon        | PyPy   | Shed Skin    | Typon                           |
|-------------------------------|---------------------|----------------------------------------------|--------------|--------|--------------|---------------------------------|
| Status                        | Mature              | Mature                                       | Mature       | Mature | Experimental | Experimental                    |
| Platform support              | Cross               | Cross                                        | Cross + WASM | Cross  | ?            | Linux only (relies on io-uring) |
| Native codegen                | AOT                 | n/a (AOT + Python)                           | AOT          | JIT    | AOT          | AOT                             |
| Required type annotations     | Function signatures | No                                           | ?            | No     | No           | Class fields                    |
| Type inference                | ? (unclear)         |                                              |              | n/a    | One-way      | Bidirectional (Hindley-Milner)  |
| Native execution              | ✓                   | Partial -- calls CPython API for many things | ✓            |        | ✓            | ✓                               |
| First-class types             |                     | ✓                                            |              | ✓      |              | ✓                               |
| Closures, bound methods       | ✓                   | ✓                                            | ✓            | ✓      |              | ✓                               |
| Interoperability with CPython | ✓                   | ✓                                            | ✓            | ✓      |              | ✓                               |
| Async, I/O coroutines         | Not for `cdef`      | ✓                                            |              | ✓      |              | ✓                               |

(non-exhaustive table, incomplete)